#  Copyright © 2022 Ingram Micro Inc. All rights reserved.
import multiprocessing
import signal

from django.core.management.base import BaseCommand, CommandError

from dj_cqrs.registries import ReplicaRegistry
from dj_cqrs.transport import current_transport


class WorkersManager:

    def __init__(self, options, transport, consume_kwargs):
        self.pool = []
        self.options = options
        self.transport = transport
        self.consume_kwargs = consume_kwargs

    def start(self):
        for i in range(self.options['workers'] or 1):
            process = multiprocessing.Process(
                name=f'cqrs-consumer-{i}',
                target=self.transport.consume,
                kwargs=self.consume_kwargs,
            )
            self.pool.append(process)
            process.start()

        for process in self.pool:
            process.join()

    def terminate(self, *args, **kwargs):
        while self.pool:
            p = self.pool.pop()
            p.terminate()
            p.join()

    def reload(self, *args, **kwargs):
        self.terminate()
        self.start()


class Command(BaseCommand):
    help = 'Starts CQRS worker, which consumes messages from message queue.'

    def add_arguments(self, parser):
        parser.add_argument('--workers', '-w', help='Number of workers', type=int, default=0)
        parser.add_argument(
            '--cqrs-id',
            '-cid',
            nargs='*',
            type=str,
            help='Choose model(s) by CQRS_ID for consuming',
        )
        parser.add_argument(
            '--reload', '-r', help='Enable reload signal SIGHUP', action='store_true',
        )

    def handle(self, *args, **options):
        if not options['workers'] and not options['reload']:
            current_transport.consume(**self.get_consume_kwargs(options))
            return

        self.start_workers_pool(options)

    def start_workers_pool(self, options):
        workers_manager = WorkersManager(
            options, current_transport, self.get_consume_kwargs(options),
        )
        if options['reload']:
            try:
                multiprocessing.set_start_method('spawn')
            except RuntimeError:
                pass

            signal.signal(signal.SIGHUP, workers_manager.reload)

        workers_manager.start()

    def get_consume_kwargs(self, options):
        consume_kwargs = {}
        if options.get('cqrs_id'):
            cqrs_ids = set()
            for cqrs_id in options['cqrs_id']:
                model = ReplicaRegistry.get_model_by_cqrs_id(cqrs_id)
                if not model:
                    raise CommandError('Wrong CQRS ID: {0}!'.format(cqrs_id))

                cqrs_ids.add(cqrs_id)

            consume_kwargs['cqrs_ids'] = cqrs_ids

        return consume_kwargs



#  Copyright © 2022 Ingram Micro Inc. All rights reserved.

import copy
import logging
from contextlib import ExitStack

from django.db import Error, close_old_connections, transaction

from dj_cqrs.constants import SignalType
from dj_cqrs.registries import ReplicaRegistry


logger = logging.getLogger('django-cqrs')


def consume(payload):
    """ Consumer controller.

    :param dj_cqrs.dataclasses.TransportPayload payload: Consumed payload from master service.
    """
    payload = copy.deepcopy(payload)
    return route_signal_to_replica_model(
        payload.signal_type,
        payload.cqrs_id,
        payload.instance_data,
        previous_data=payload.previous_data,
        meta=payload.meta,
    )


def route_signal_to_replica_model(
    signal_type, cqrs_id, instance_data, previous_data=None, meta=None,
):
    """ Routes signal to model method to create/update/delete replica instance.

    :param dj_cqrs.constants.SignalType signal_type: Consumed signal type.
    :param str cqrs_id: Replica model CQRS unique identifier.
    :param dict instance_data: Master model data.
    :param dict or None previous_data: Previous model data for changed tracked fields, if exists.
    :param dict or None meta: Payload metadata, if exists.
    """
    if signal_type not in (SignalType.DELETE, SignalType.SAVE, SignalType.SYNC):
        logger.error('Bad signal type "{0}" for CQRS_ID "{1}".'.format(signal_type, cqrs_id))
        return

    model_cls = ReplicaRegistry.get_model_by_cqrs_id(cqrs_id)
    if model_cls:
        db_is_needed = not model_cls.CQRS_NO_DB_OPERATIONS
        if db_is_needed:
            close_old_connections()

        is_meta_supported = model_cls.CQRS_META
        try:
            with transaction.atomic(savepoint=False) if db_is_needed else ExitStack():
                if signal_type == SignalType.DELETE:
                    if is_meta_supported:
                        return model_cls.cqrs_delete(instance_data, meta=meta)

                    return model_cls.cqrs_delete(instance_data)

                f_kw = {'previous_data': previous_data}
                if is_meta_supported:
                    f_kw['meta'] = meta

                if signal_type == SignalType.SAVE:
                    return model_cls.cqrs_save(instance_data, **f_kw)

                if signal_type == SignalType.SYNC:
                    f_kw['sync'] = True
                    return model_cls.cqrs_save(instance_data, **f_kw)

        except Error as e:
            pk_value = instance_data.get(model_cls._meta.pk.name)
            cqrs_revision = instance_data.get('cqrs_revision')

            logger.error(
                '{0}\nCQRS {1} error: pk = {2}, cqrs_revision = {3} ({4}).'.format(
                    str(e), signal_type, pk_value, cqrs_revision, model_cls.CQRS_ID,
                ),
            )
